package net.wanji.opt.service.impl;

import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import net.wanji.common.enums.CrossStatusEnum;
import net.wanji.common.framework.Constants;
import net.wanji.common.framework.rest.JsonViewObject;
import net.wanji.common.utils.tool.BeanListUtils;
import net.wanji.databus.dao.entity.RidInfoEntity;
import net.wanji.databus.dao.mapper.CrossDirDataRealtimeMapper;
import net.wanji.databus.dao.mapper.CrossTurnDataRealtimeMapper;
import net.wanji.databus.dto.CrossDirInfoDTO;
import net.wanji.databus.dto.CrossTurnInfoDTO;
import net.wanji.databus.entity.develop.servicedevelop.develop.StatusCodeEnum;
import net.wanji.databus.po.CrossDataRealtimePO;
import net.wanji.databus.po.CrossDirDataRealtimePO;
import net.wanji.databus.po.CrossTurnDataRealtimePO;
import net.wanji.databus.vo.SchemeSendVO;
import net.wanji.databus.vo.SignalStatusVO;
import net.wanji.feign.service.UtcFeignClients;
import net.wanji.opt.cache.CrossDirTurnPhaseCache;
import net.wanji.opt.cache.CrossRidInfoCache;
import net.wanji.opt.common.exception.OptServiceException;
import net.wanji.opt.dao.mapper.CrossSchemeOptLogMapper;
import net.wanji.opt.dto.CrossPhaseDTO;
import net.wanji.opt.dto.CrossTurnDataRealtimeDTO;
import net.wanji.opt.po.base.CrossPhasePO;
import net.wanji.opt.po.base.CrossSchemeOptLogPO;
import net.wanji.opt.po.base.CrossSchemePO;
import net.wanji.opt.po.strategy.SceneStrategyIdeaPO;
import net.wanji.opt.service.CrossOptimizeService;
import net.wanji.opt.service.strategy.SceneService;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.CollectionUtils;

import java.time.LocalDateTime;
import java.time.ZoneOffset;
import java.util.*;
import java.util.stream.Collectors;

import static net.wanji.opt.cache.CrossPhaseInfoCache.getCrossPhaseCacheByCrossId;
import static net.wanji.opt.cache.CrossSchemeInfoCache.geteSchemePOCache;

/**
 * @author hfx
 * @date 2023/1/10 15:15
 * @desc CrossOptimizeServiceImpl
 */
@Service
@RequiredArgsConstructor
@Slf4j
public class CrossOptimizeServiceImpl implements CrossOptimizeService {

    private final CrossSchemeOptLogMapper crossSchemeOptLogMapper;
    private final UtcFeignClients utcFeignClients;
    private final CrossDirTurnPhaseCache crossDirTurnPhaseCache;
    private final CrossRidInfoCache crossRidInfoCache;
    private final SceneService sceneService;
    private final CrossTurnDataRealtimeMapper crossTurnDataRealtimeMapper;
    private final CrossDirDataRealtimeMapper crossDirDataRealtimeMapper;

    @Value("${crossOptParam.maxVehheadDist}")
    private Double maxVehheadDist;
    @Value("${crossOptParam.minVehheadDist}")
    private Double minVehheadDist;
    @Value("${crossOptParam.defaultVehheadDist}")
    private Double defaultVehheadDist;
    @Value("${crossOptParam.maxVehheadTime}")
    private Double maxVehheadTime;
    @Value("${crossOptParam.minVehheadTime}")
    private Double minVehheadTime;
    @Value("${crossOptParam.defaultVehheadTime}")
    private Double defaultVehheadTime;

    static Set<String> CROSS_OPT = new HashSet<>();   // 记录已优化的路口

    @Override
    @Transactional
    public String realtimeOptimize(List<CrossDataRealtimePO> abnormalCrossList, List<CrossDataRealtimePO> crossDataRealtimePOList) {

        try {
            // todo 当前路口不是信控路口，需要跳过
//            utcFeignClients.listCrossInfo()
            // todo 系统管理，路口管理，信控标志关闭，优化下发也要关闭

            // 相位配时信息，key: 路口编号_方向类型_转向类型，value: 相位配时信息
            crossDirTurnPhaseCache.refresh();
            Map<String, CrossPhaseDTO> phaseMap = crossDirTurnPhaseCache.getPhaseDirTurnMap();
            setTurnList(abnormalCrossList, phaseMap);
            setTurnList(crossDataRealtimePOList, phaseMap);

            // 获取所有信控路口转向实时数据，包括异常路口和正常路口
//            Map<String, List<CrossTurnDataRealtimeDTO>> turnDataRealtimeMap = listTurnDataRealtime(crossDataRealtimePOList);
            List<CrossTurnDataRealtimePO> crossTurnDataRealtimePOS = crossTurnDataRealtimeMapper.selectList(new LambdaQueryWrapper<CrossTurnDataRealtimePO>());
            ArrayList<CrossTurnDataRealtimeDTO> turnDataRealtimeList = new ArrayList<>();
            BeanListUtils.populateList(crossTurnDataRealtimePOS, turnDataRealtimeList, CrossTurnDataRealtimeDTO.class);
            Map<String, List<CrossTurnDataRealtimeDTO>> turnDataRealtimeMap = turnDataRealtimeList.stream().collect(Collectors.groupingBy(CrossTurnDataRealtimeDTO::getCrossId));

            // 恢复优化路口原始方案:上一批次优化后，路口正常需要恢复原始方案
            // todo 方案运行时间，恢复时当前时间是否在远方案运行时间内，否则phaseMap数据不准确
            restoreOptCrossOriSchema(abnormalCrossList, phaseMap);

            Map<Integer, List<CrossDataRealtimePO>> crossDataMap = abnormalCrossList.stream().collect(Collectors.groupingBy(CrossDataRealtimePO::getStatus));
            abnormalCrossList = crossDataMap.get(CrossStatusEnum.SPILLOVER.getCode());  // 溢出
            if (abnormalCrossList != null && !abnormalCrossList.isEmpty()) {
                spilloverOpt(abnormalCrossList, turnDataRealtimeMap, phaseMap);
            }
            abnormalCrossList = crossDataMap.get(CrossStatusEnum.CONGESTION.getCode()); // 拥堵
            if (abnormalCrossList != null && !abnormalCrossList.isEmpty()) {
                congestionOpt(abnormalCrossList, turnDataRealtimeMap, phaseMap);
            }
            abnormalCrossList = crossDataMap.get(CrossStatusEnum.UNBALANCE.getCode());  // 失衡
            if (abnormalCrossList != null && !abnormalCrossList.isEmpty()) {
                unbalanceOpt(abnormalCrossList, turnDataRealtimeMap, phaseMap);
            }
        } catch (Exception e) {
            e.printStackTrace();
            log.error("实时优化", e.getMessage());
        }
        return null;
    }

    /**
     * 路口溢出优化
     *
     * @param abnormalCrossList
     * @param turnDataRealtimeMap
     * @param phaseMap
     */
    private void spilloverOpt(List<CrossDataRealtimePO> abnormalCrossList, Map<String, List<CrossTurnDataRealtimeDTO>> turnDataRealtimeMap,
                              Map<String, CrossPhaseDTO> phaseMap) throws Exception {

        // todo 获取溢出场景当前执行的策略方法的指标数据，通过指标数据进行下发方案
        SceneStrategyIdeaPO sceneStrategyIdeaPO = sceneService.selectIdBySceneNum(CrossStatusEnum.SPILLOVER.getCode());
        // 溢出路口集合
        Map<String, List<CrossDataRealtimePO>> congestCrossMap = abnormalCrossList.stream().collect(Collectors.groupingBy(CrossDataRealtimePO::getCrossId));

        String crossId;
        for (CrossDataRealtimePO cross : abnormalCrossList) {

            crossId = cross.getCrossId();
            // 判断信号机是否在线 todo 是否某些故障也不需要优化
            // 判断当前路口是否存在特殊控制操作
            if (isOffLineOrSpecialControlMode(crossId)) {
                continue;
            }

            // 判断当前路口是否已优化
            if (CROSS_OPT.contains(crossId)) {
                continue;
            }
            // 判断路口是否是绿波
            // 路口拥堵进口方向，dir
            List<Integer> optCrossDirList = getOptCrossInDirList(crossId);

            // 溢出当前路口方向转向减少时间 10s
            // 计算路口各转向所需的通行时长（秒）
            List<CrossTurnDataRealtimeDTO> crossTurnDataRealtimeDTOS = turnDataRealtimeMap.get(crossId);
            for (Integer dir : optCrossDirList) {
                for (CrossTurnDataRealtimeDTO crossTurnDataRealtimeDTO : crossTurnDataRealtimeDTOS) {
                    Integer inDir = crossTurnDataRealtimeDTO.getInDir();
                    if (Objects.equals(dir, inDir)) {
                        crossTurnDataRealtimeDTO.setPassTime(-10.0);
                    }
                }
            }
            // 溢出优化当前路口 直行
            doExecuteCrossOpt(crossTurnDataRealtimeDTOS, phaseMap, crossId, "3");

            // 下游路口通过可排队长度计算可加可减时间优化
            // 获取下游路口
            List<RidInfoEntity> ridInfoEntities = crossRidInfoCache.getRidInfoListByCrossId(crossId, Constants.SystemParam.NULL);
            // 不存在下游路口，不做处理
            if (CollectionUtils.isEmpty(ridInfoEntities)) {
                continue;
            }
            for (Integer dir : optCrossDirList) {
                for (RidInfoEntity ridInfoEntity : ridInfoEntities) {
                    Integer inDir = ridInfoEntity.getInDir();
                    if (!congestCrossMap.containsKey(ridInfoEntity.getEndCrossId()) && Objects.equals(inDir, dir)) {
                        String endCrossId = ridInfoEntity.getEndCrossId();
                        // 判断当前路口是否已优化
                        if (CROSS_OPT.contains(endCrossId)) {
                            continue;
                        }
                        List<CrossTurnDataRealtimeDTO> endcrossTurnDataRealtimeDTOS = turnDataRealtimeMap.get(endCrossId);
                        if (CollectionUtils.isEmpty(endcrossTurnDataRealtimeDTOS)) {
                            log.error("溢出路口：{}，优化其下游路口：{}时，路口转向数据为空", crossId, ridInfoEntity.getEndCrossId());
                            throw new OptServiceException("实时监控->溢出路口方案优化异常，优化路口转向数据异常");
                        }
                        endcrossTurnDataRealtimeDTOS.forEach(item -> item.setPassTime(calPassTime(item)));
                        // 进行优化下发
                        doExecuteCrossOpt(crossTurnDataRealtimeDTOS, phaseMap, endCrossId, "3");
                        // 记录已优化的路口
                        CROSS_OPT.add(endCrossId);
                    }
                }
            }
            // 记录已优化的路口
            CROSS_OPT.add(crossId);
        }
    }

    /**
     * 路口拥堵优化
     *
     * @param abnormalCrossList
     * @param turnDataRealtimeMap
     * @param phaseMap
     */
    private void congestionOpt(List<CrossDataRealtimePO> abnormalCrossList, Map<String, List<CrossTurnDataRealtimeDTO>> turnDataRealtimeMap,
                               Map<String, CrossPhaseDTO> phaseMap) throws Exception {

        // 获取拥堵场景当前执行的策略方法的指标数据，通过指标数据进行下发方案
        SceneStrategyIdeaPO sceneStrategyIdeaPO = sceneService.selectIdBySceneNum(CrossStatusEnum.CONGESTION.getCode());

        // 拥堵路口集合
        Map<String, List<CrossDataRealtimePO>> congestCrossMap = abnormalCrossList.stream().collect(Collectors.groupingBy(CrossDataRealtimePO::getCrossId));

        String crossId;
        for (CrossDataRealtimePO cross : abnormalCrossList) {

            crossId = cross.getCrossId();
            // 判断信号机是否在线 todo 是否某些故障也不需要优化
            // 判断当前路口是否存在特殊控制操作
            if (isOffLineOrSpecialControlMode(crossId)) {
                continue;
            }

            // 路口拥堵方向转向，dir_turn
            List<Integer> optCrossDirList = getOptCrossInDirList(crossId);

            // 判断当前路口是否已优化
            if (CROSS_OPT.contains(crossId)) {
                continue;
            }
            // 判断路口是否是绿波
            // 获取上游路口
            List<RidInfoEntity> ridInfoEntities = crossRidInfoCache.getRidInfoListByCrossId(Constants.SystemParam.NULL, crossId);
            // 不存在上游路口，不做处理
            if (CollectionUtils.isEmpty(ridInfoEntities)) {
                continue;
            }
            // 当前路口拥堵方向，优化当前路口方向的上游路口的直行和左转方向
            for (Integer intCongestionDir : optCrossDirList) {
                // 拥堵路口对上游路口进       行优化
                for (RidInfoEntity ridInfoEntity : ridInfoEntities) {
                    String startCrossId = ridInfoEntity.getStartCrossId();

                    // 上游路口时拥堵路口时，不进行优化或者判断当前路口是否已优化
                    if (congestCrossMap.containsKey(startCrossId) || CROSS_OPT.contains(startCrossId)) {
                        continue;
                    }

                    Integer inDir = ridInfoEntity.getInDir();
                    // 优化当前开始路口的此方向直行，和逆时针方向左转
                    if (Objects.equals(inDir, intCongestionDir)) {
                        // 计算上游路口各转向所需的通行时长（秒）  此方向和逆时针左转可可减时间 10s
                        List<CrossTurnDataRealtimeDTO> crossTurnDataRealtimeDTOS = turnDataRealtimeMap.get(startCrossId);
                        if (!CollectionUtils.isEmpty(crossTurnDataRealtimeDTOS)) {
                            for (CrossTurnDataRealtimeDTO crossTurnDataRealtimeDTO : crossTurnDataRealtimeDTOS) {
                                Integer startInDir = crossTurnDataRealtimeDTO.getInDir();
                                String turnType = crossTurnDataRealtimeDTO.getTurnType();
                                if (Objects.equals(startInDir, intCongestionDir)) {
                                    crossTurnDataRealtimeDTO.setPassTime(-10.0);
                                }
                                if (Objects.equals(startInDir, (intCongestionDir + 2) == 9 ? 1 : intCongestionDir + 2) && Objects.equals("l", turnType)) {
                                    crossTurnDataRealtimeDTO.setPassTime(-10.0);
                                }
                            }
                            // 进行优化下发
                            doExecuteCrossOpt(crossTurnDataRealtimeDTOS, phaseMap, startCrossId, "2");

                            // 记录已优化的路口
                            CROSS_OPT.add(crossId);
                        } else {
                            log.error("当前拥堵路口： {}， 实时监控->优化上游路口： {}时，未获取到上游路口的实时转向数据！", crossId, startCrossId);
                        }
                    }
                }
            }
        }
    }

    /**
     * 路口失衡优化
     *
     * @param abnormalCrossList   失衡路口实时数据
     * @param turnDataRealtimeMap 路口转向实时数据
     * @param phaseMap            路口相位配时数据
     */
    private void unbalanceOpt(List<CrossDataRealtimePO> abnormalCrossList, Map<String, List<CrossTurnDataRealtimeDTO>> turnDataRealtimeMap,
                              Map<String, CrossPhaseDTO> phaseMap) throws Exception {

        // 获取失衡场景当前执行的策略方法的指标数据，通过指标数据进行下发方案
        SceneStrategyIdeaPO sceneStrategyIdeaPO = sceneService.selectIdBySceneNum(CrossStatusEnum.CONGESTION.getCode());

        String crossId;
        for (CrossDataRealtimePO cross : abnormalCrossList) {

            crossId = cross.getCrossId();
            // 判断信号机是否在线 todo 是否某些故障也不需要优化
            // 判断当前路口是否存在特殊控制操作
            if (isOffLineOrSpecialControlMode(crossId)) {
                continue;
            }
            // 判断当前路口是否已优化
            if (CROSS_OPT.contains(crossId)) {
                continue;
            }
            // 判断路口是否是绿波

            // 计算路口各转向所需的通行时长（秒）
            List<CrossTurnDataRealtimeDTO> crossTurnDataRealtimeDTOS = turnDataRealtimeMap.get(crossId);
            crossTurnDataRealtimeDTOS.forEach(item -> item.setPassTime(calPassTime(item)));
            // 执行路口优化
            doExecuteCrossOpt(crossTurnDataRealtimeDTOS, phaseMap, crossId, "1");

            // 记录已优化的路口
            CROSS_OPT.add(crossId);
        }
    }

    private List<Integer> getOptCrossInDirList(String crossId) {
        LambdaQueryWrapper<CrossDirDataRealtimePO> dirWrapper = new LambdaQueryWrapper<>();
        dirWrapper.eq(CrossDirDataRealtimePO::getCrossId, crossId);
        dirWrapper.eq(CrossDirDataRealtimePO::getInOutType, 1);
        List<CrossDirDataRealtimePO> crossDirDataRealtimePOS = crossDirDataRealtimeMapper.selectList(dirWrapper);
        List<Integer> optCrossDirList = crossDirDataRealtimePOS.stream().map(CrossDirDataRealtimePO::getDirType).collect(Collectors.toList());
        return optCrossDirList;
    }

    /**
     * 设置异常数据转向数据
     *
     * @param abnormalCrossList
     * @param phaseMap
     */
    private static void setTurnList(List<CrossDataRealtimePO> abnormalCrossList, Map<String, CrossPhaseDTO> phaseMap) {
        for (CrossDataRealtimePO crossDataRealtimeDTO : abnormalCrossList) {
            List<CrossTurnInfoDTO> turnList = new ArrayList<>();
            List<CrossDirInfoDTO> dirList = new ArrayList<>();
            for (Map.Entry<String, CrossPhaseDTO> entry : phaseMap.entrySet()) {
                String crossDirTurn = entry.getKey();
                String[] split = crossDirTurn.split(Constants.SystemParam.SEPARATOR_UNDER_LINE);
                String crossId = split[0];
                String dir = split[1];
                String turn = split[2];

                CrossTurnInfoDTO crossTurnInfoDTO = new CrossTurnInfoDTO();
                CrossDirInfoDTO crossDirInfoDTO = new CrossDirInfoDTO();
                if (StringUtils.equals(crossId, crossDataRealtimeDTO.getCrossId())) {

                    crossDirInfoDTO.setCrossId(crossId);
                    crossDirInfoDTO.setInOutType(1);
                    crossDirInfoDTO.setDirType(Integer.valueOf(dir));
                    dirList.add(crossDirInfoDTO);

                    crossTurnInfoDTO.setTurnType(turn);
                    crossTurnInfoDTO.setInDir(Integer.valueOf(dir));
                    crossTurnInfoDTO.setCrossId(crossId);
                    turnList.add(crossTurnInfoDTO);
                }
            }
        }
    }

    /**
     * 通过第二批次列表，判断需要恢复的优化路口原始方案
     *
     * @param abnormalCrossList
     */
    private void restoreOptCrossOriSchema(List<CrossDataRealtimePO> abnormalCrossList, Map<String, CrossPhaseDTO> phaseMap) throws Exception {
        List<String> abnormalCrossIdList = abnormalCrossList.stream().map(CrossDataRealtimePO::getCrossId).collect(Collectors.toList());
        for (String optCrossId : CROSS_OPT) {
            if (!abnormalCrossIdList.contains(optCrossId)) {
                // 下发原始方案
                SchemeSendVO schemeSendVO = getSchemeSendVO(optCrossId, phaseMap, Collections.EMPTY_MAP);
                JsonViewObject jsonViewObject = utcFeignClients.schemeSend(schemeSendVO);
                if (Objects.isNull(jsonViewObject) || jsonViewObject.getCode() != 200) {
                    log.error("实时监控->路口优化方案下发远程调用异常！", jsonViewObject);
                    throw new OptServiceException("实时监控->恢复原始方案调用UTC服务异常");
                }
                CROSS_OPT.remove(optCrossId);
            }
        }
    }


    /**
     * 执行路口优化逻辑
     * 计算路口相位优化时长并下发方案
     *
     * @param crossTurnDataRealtimeDTOS
     * @param phaseMap
     * @param crossId
     * @throws Exception
     */
    private void doExecuteCrossOpt(List<CrossTurnDataRealtimeDTO> crossTurnDataRealtimeDTOS, Map<String, CrossPhaseDTO> phaseMap, String crossId, String optType) throws Exception {
        // 计算路口各相位调整时长（秒）
        Map<String, Integer> phaseTimeOptResultMap = new HashMap<>();
        Map<String, Integer> phaseTimeOffset = getTimeOffsetPhaseMap(phaseMap, crossTurnDataRealtimeDTOS);
        // 是否需要拆分相位
        if (isSplitPhase(phaseTimeOffset)) {

        } else {
            phaseTimeOptResultMap = getPhaseTimeOptMap(phaseTimeOffset);
        }
        if (!phaseTimeOptResultMap.isEmpty()) {
            // 修改原始方案绿灯时间为优化后时间，方案下发
            SchemeSendVO schemeSendVO = getSchemeSendVO(crossId, phaseMap, phaseTimeOptResultMap);
            JsonViewObject jsonViewObject = utcFeignClients.schemeSend(schemeSendVO);

            //todo  test下发成功
            jsonViewObject.success();
            if (Objects.isNull(jsonViewObject) || jsonViewObject.getCode() != 200) {
                log.error("实时监控->路口优化方案下发远程调用异常！", jsonViewObject);
                throw new OptServiceException("实时监控->下发优化方案调用UTC服务异常");
            }

            // 将优化记录插入方案优化记录表
            insertCrossSchemeOptLog(phaseTimeOptResultMap, phaseMap, optType);
        }
    }

    /**
     * 将优化记录数据[按路口相位转向]插入到优化记录表这种
     *
     * @param phaseTimeOptResultMap
     * @param phaseMap
     */
    private void insertCrossSchemeOptLog(Map<String, Integer> phaseTimeOptResultMap, Map<String, CrossPhaseDTO> phaseMap, String optType) {
        List<CrossSchemeOptLogPO> insertList = new ArrayList<>();
        // 数据批次
        long dataBatchTimeLong = LocalDateTime.now().toEpochSecond(ZoneOffset.of("+8"));
        for (Map.Entry<String, CrossPhaseDTO> item : phaseMap.entrySet()) {
            CrossSchemeOptLogPO crossSchemeOptLogPO = new CrossSchemeOptLogPO();

            CrossPhaseDTO crossPhaseDTO = item.getValue();
            String phaseNo = crossPhaseDTO.getPhaseNo();
            for (Map.Entry<String, Integer> entry : phaseTimeOptResultMap.entrySet()) {
                if (Objects.equals(phaseNo, entry.getKey())) {
                    BeanUtils.copyProperties(crossPhaseDTO, crossSchemeOptLogPO);
                    crossSchemeOptLogPO.setSchemeId(crossPhaseDTO.getSchemeId());
                    crossSchemeOptLogPO.setPhaseOrderId(crossPhaseDTO.getSort());
                    crossSchemeOptLogPO.setStartTime(new Date());
                    crossSchemeOptLogPO.setOptType("1");// 自动优化

                    crossSchemeOptLogPO.setOptType(optType);
                    crossSchemeOptLogPO.setOptResult("1");
                    crossSchemeOptLogPO.setOptResultDesc(StatusCodeEnum.STATUS_00200.getDetail());
                    crossSchemeOptLogPO.setDataBatchTime((int) dataBatchTimeLong);
                    // todo 测试数据
                    crossSchemeOptLogPO.setDataExtend("{\"sceneName\": \"测试JSON\",\"strategyName\": \"测试JSON\",\"ideaName\": \"测试JSON\"}");// todo 需要通过场景策略方法

                    crossSchemeOptLogPO.setOptTime(entry.getValue());
                    crossSchemeOptLogPO.setOriGreenTime(crossPhaseDTO.getGreenTime());
                    insertList.add(crossSchemeOptLogPO);
                }
            }
        }
        crossSchemeOptLogMapper.insertBatch(insertList);
    }

    /**
     * 构建下发的优化方案
     * 如果phaseTimeOptResultMap=null,下发的原始方案
     * 否则原始方案和相位优化时间参数phaseTimeOptResultMap构建下发参数
     *
     * @param crossId
     * @param phaseMap
     * @param phaseTimeOptResultMap
     * @return
     */
    public static SchemeSendVO getSchemeSendVO(String crossId, Map<String, CrossPhaseDTO> phaseMap, Map<String, Integer> phaseTimeOptResultMap) throws Exception {
        String key = String.join(Constants.SystemParam.SEPARATOR_UNDER_LINE, crossId, "1", "s");
        CrossPhaseDTO crossPhaseDTO = phaseMap.get(key);
        if (Objects.isNull(crossPhaseDTO)) {
            log.error("当前路口：{}未获取到信号机方案信息，无法进行方案下发优化", crossId);
            throw new OptServiceException("当前路口不能获取到运行方案相位信息!");
        }
        Integer schemeId = crossPhaseDTO.getSchemeId();

        CrossSchemePO crossSchemePO = geteSchemePOCache(String.join(Constants.SystemParam.SEPARATOR_UNDER_LINE, crossId, String.valueOf(schemeId)));
        Map<Integer, List<CrossPhasePO>> ringNoPhasePOMap = getCrossPhaseCacheByCrossId(crossId).stream().collect(Collectors.groupingBy(CrossPhasePO::getRingNo));

        SchemeSendVO schemeSendVO = new SchemeSendVO();
        List<SchemeSendVO.Pattern> patternList = new ArrayList<>();
        if (Objects.nonNull(crossSchemePO)) {
            SchemeSendVO.Pattern pattern = new SchemeSendVO.Pattern();
            pattern.setPatternName("实时监测优化方案".concat(crossSchemePO.getSchemeNo()));
            pattern.setPatternNo(String.valueOf(crossSchemePO.getId()));
            pattern.setCycle(String.valueOf(crossSchemePO.getCycle()));
            pattern.setOffset(String.valueOf(crossSchemePO.getOffset()));
            pattern.setCoordPhase(String.valueOf(crossSchemePO.getCoordPhase()));

            List<SchemeSendVO.Pattern.Ring> rings = new ArrayList<>();
            ringNoPhasePOMap.entrySet().forEach(entry -> {
                SchemeSendVO.Pattern.Ring ring = new SchemeSendVO.Pattern.Ring();
                List<SchemeSendVO.Pattern.Ring.Phase> phaseList = new ArrayList<>();
                List<CrossPhasePO> value = entry.getValue();
                value.forEach(crossPhasePO -> {
                    SchemeSendVO.Pattern.Ring.Phase phase = new SchemeSendVO.Pattern.Ring.Phase();
                    phase.setPhaseNo(crossPhasePO.getPhaseNo());
                    phase.setPhaseName(crossPhasePO.getName());
                    phase.setSort(String.valueOf(crossPhasePO.getSort()));
                    phase.setControlMode(String.valueOf(crossPhasePO.getControlMode()));
                    phase.setMinGreenTime(String.valueOf(crossPhasePO.getMinGreenTime()));
                    phase.setMaxGreenTime(String.valueOf(crossPhasePO.getMaxGreenTime()));

                    // 可优化时间
                    Integer currentPhaseOffset = getCurrentPhaseOptOffsetTime(phaseTimeOptResultMap, crossPhasePO.getPhaseNo());
                    phase.setPhaseTime(String.valueOf(crossPhasePO.getPhaseTime() + currentPhaseOffset));
                    phase.setGreenTime(String.valueOf(crossPhasePO.getGreenTime() + currentPhaseOffset));
                    phase.setGreenFlashTime(String.valueOf(crossPhasePO.getGreenFlashTime()));
                    phase.setPedFlashTime(String.valueOf(crossPhasePO.getPedFlashTime()));
                    phase.setYellowTime(String.valueOf(crossPhasePO.getYellowTime()));
                    phase.setRedTime(String.valueOf(crossPhasePO.getRedTime()));
                    phaseList.add(phase);
                });
                ring.setRingNo(String.valueOf(entry.getKey()));
                ring.setPhaseList(phaseList);
                rings.add(ring);
            });
            pattern.setRings(rings);
            patternList.add(pattern);
        }

        schemeSendVO.setCrossCode(crossId);
        schemeSendVO.setPatternList(patternList);
        return schemeSendVO;
    }


    /**
     * 获取当前相位的可优化时间，默认0
     *
     * @param phaseTimeOptResultMap
     * @param phaseNo
     * @return
     */
    private static int getCurrentPhaseOptOffsetTime(Map<String, Integer> phaseTimeOptResultMap, String phaseNo) {
        if (Objects.nonNull(phaseTimeOptResultMap) && !phaseTimeOptResultMap.isEmpty()) {
            for (Map.Entry<String, Integer> phaseEntry : phaseTimeOptResultMap.entrySet()) {
                if (StringUtils.equals(phaseNo, phaseEntry.getKey())) {
                    Integer optTime = phaseTimeOptResultMap.get(phaseNo);
                    return optTime;
                }
            }
        }
        return 0;
    }

    /**
     * 在最大最小绿范围内，通过转向优化计算相位可优化时间
     * key:相位号 value:offsetTime
     *
     * @param timeOffsetPhaseMap
     * @return
     */
    private static Map<String, Integer> getPhaseTimeOptMap(Map<String, Integer> timeOffsetPhaseMap) {
        HashMap<String, Integer> phaseTimeOptResultMap = new HashMap<>();
        // key:相位号 value:最大可加减时间
        Map<String, Integer> phaseOffsetTimeSetMap = new HashMap<>();
        for (Map.Entry<String, Integer> entry : timeOffsetPhaseMap.entrySet()) {
            Integer offsetTime = entry.getValue() == null ? 0 : entry.getValue();
            String phaseNo = entry.getKey();
            if (phaseOffsetTimeSetMap.containsKey(phaseNo)) {
                Integer currentOffsetTime = phaseOffsetTimeSetMap.get(phaseNo);
                if (Math.abs(offsetTime) > Math.abs(currentOffsetTime)) {
                    phaseOffsetTimeSetMap.put(phaseNo, offsetTime);
                } else {
                    phaseOffsetTimeSetMap.put(phaseNo, currentOffsetTime);
                }
            } else {
                phaseOffsetTimeSetMap.put(phaseNo, offsetTime);
            }
        }
        List<Integer> subTimeList = phaseOffsetTimeSetMap.entrySet().stream().map(entry -> entry.getValue()).collect(Collectors.toList()).stream().filter(i -> i < 0).collect(Collectors.toList());
        List<Integer> addTimeList = phaseOffsetTimeSetMap.entrySet().stream().map(entry -> entry.getValue()).collect(Collectors.toList()).stream().filter(i -> i > 0).collect(Collectors.toList());
        int subTimeSum = subTimeList.stream().mapToInt(Integer::intValue).sum();
        int addTimeSum = addTimeList.stream().mapToInt(Integer::intValue).sum();
        int countOffsetAdd = Math.abs(subTimeSum) >= addTimeSum ? addTimeSum : Math.abs(subTimeSum);
        int countOffsetSub = -countOffsetAdd;
        // 可加可减时间为0，不需要计算相位时间
        if (countOffsetAdd == 0) {
            return Collections.EMPTY_MAP;
        }
        for (Map.Entry<String, Integer> entry : phaseOffsetTimeSetMap.entrySet()) {
            Integer offsetTime = entry.getValue();
            String phaseNo = entry.getKey();
            if (countOffsetAdd <= 0 && countOffsetSub >= 0) {
                break;
            }
            // 可减时间有可加时间
            if (offsetTime < 0 && countOffsetSub < 0) {
                countOffsetSub -= offsetTime;
                if (countOffsetSub <= 0) {
                    phaseTimeOptResultMap.put(phaseNo, offsetTime);
                } else {
                    phaseTimeOptResultMap.put(phaseNo, -countOffsetSub);
                }
            }
            // 可加时间有可减时间
            if (offsetTime > 0 && countOffsetAdd > 0) {
                countOffsetAdd -= offsetTime;
                if (countOffsetAdd >= 0) {
                    phaseTimeOptResultMap.put(phaseNo, offsetTime);
                } else {
                    phaseTimeOptResultMap.put(phaseNo, countOffsetAdd);
                }
            }
        }

        return phaseTimeOptResultMap;
    }

    /**
     * 是否需要拆分相位灯组
     * 同一相位内不同转向同时存在可加可减，需要拆分相位
     *
     * @param timeOffsetPhaseMap
     * @return
     */
    private boolean isSplitPhase(Map<String, Integer> timeOffsetPhaseMap) {
        HashMap<String, Boolean> phaseOffsetMap = new HashMap<>();
        for (Map.Entry<String, Integer> entry : timeOffsetPhaseMap.entrySet()) {
            Boolean gtZore = entry.getValue() >= 0 ? Boolean.TRUE : Boolean.FALSE;
            String phaseNo = entry.getKey();
            if (phaseOffsetMap.containsKey(phaseNo) && !phaseOffsetMap.containsValue(gtZore)) {
                return true;
            }
        }
        return false;
    }

    /**
     * 计算路口各相位调整时长（秒）
     * 相位所需总放行时长 = sum(相位所需放行时长) - 总放行时长（周期）
     * 相位可减时长 = 相位有效绿灯时长 - 相位所需时长，同时“相位剩余有效绿灯时长”必须介于最小绿时长与最大绿时长之间
     * 相位可加时长 = 相位所需时长 - 有效绿灯时长
     * 相位可用总时长 = sum(相位可减时长)
     * 相位优化时长 = 原相位时长 + 相位调整时长（相位可减时长或相位可加时长）
     *
     * @param phaseMap
     * @param crossTurnDataRealtimeDTOS
     * @return key:相位编号  value:相位可加减时间
     */
    private Map<String, Integer> getTimeOffsetPhaseMap(Map<String, CrossPhaseDTO> phaseMap, List<CrossTurnDataRealtimeDTO> crossTurnDataRealtimeDTOS) {
        Map<String, Integer> timeOffsetPhaseMap = new HashMap<>();
        crossTurnDataRealtimeDTOS.forEach(item -> {
            String key = String.join(Constants.SystemParam.SEPARATOR_UNDER_LINE, item.getCrossId(), String.valueOf(item.getInDir()), String.valueOf(item.getTurnType()));
            Double passTime = item.getPassTime();

            phaseMap.entrySet().forEach(entry -> {
                String crossIdDirTurn = entry.getKey();
                CrossPhaseDTO crossPhaseDTO = entry.getValue();
                String phaseNo = crossPhaseDTO.getPhaseNo();

                if (StringUtils.equals(key, crossIdDirTurn)) {
                    Integer realOptGreenTime = getRealOptGreenTime(passTime, crossPhaseDTO);
                    Integer greenTimeOffset = realOptGreenTime - crossPhaseDTO.getGreenTime() - crossPhaseDTO.getGreenFlashTime();
                    Integer realGreenTimeOffset = 0;
                    if (!timeOffsetPhaseMap.isEmpty() && timeOffsetPhaseMap.get(phaseNo) != null) {
                        realGreenTimeOffset = timeOffsetPhaseMap.get(phaseNo);
                        timeOffsetPhaseMap.put(phaseNo, greenTimeOffset);
                        if (greenTimeOffset < 0 && realGreenTimeOffset > greenTimeOffset) {
                            timeOffsetPhaseMap.put(phaseNo, realGreenTimeOffset);
                        }
                        if (greenTimeOffset >= 0 && realGreenTimeOffset < greenTimeOffset) {
                            timeOffsetPhaseMap.put(phaseNo, realGreenTimeOffset);
                        }
                    } else {
                        timeOffsetPhaseMap.put(phaseNo, greenTimeOffset);
                    }
                }
            });
        });
        return timeOffsetPhaseMap;
    }

    private Integer getRealOptGreenTime(Double passTime, CrossPhaseDTO crossPhaseDTO) {
        Integer passTimeInt = passTime == null ? 0 : passTime.intValue();
        if (passTimeInt <= crossPhaseDTO.getMinGreenTime()) {
            passTimeInt = crossPhaseDTO.getMinGreenTime();
        }
        if (passTimeInt >= crossPhaseDTO.getMaxGreenTime()) {
            passTimeInt = crossPhaseDTO.getMaxGreenTime();
        }
        return passTimeInt;
    }


    private boolean isOffLineOrSpecialControlMode(String crossId) throws Exception {
        JsonViewObject jsonViewObject = utcFeignClients.runningStatusAlarm();
        if (Objects.isNull(jsonViewObject) || jsonViewObject.getCode() != 200) {
            log.error("路口优化获取当前路口控制模式异常！");
            throw new OptServiceException("路口优化获取当前路口控制模式异常！");
        }
        List<SignalStatusVO> signalStatusVOS = (List<SignalStatusVO>) jsonViewObject.getContent();
        List<SignalStatusVO> resultList = signalStatusVOS.stream().filter(signalStatusVO -> Objects.equals(crossId, signalStatusVO.getCrossId())).collect(Collectors.toList());
        if (!CollectionUtils.isEmpty(resultList)) {
            SignalStatusVO signalStatusVO = resultList.get(0);
            if (0 == signalStatusVO.getStatus() || 1 != signalStatusVO.getControlType()) {
                return true;
            }
        }
        return false;
    }

    /**
     * 获取所有信控路口转向实时数据
     * key 路口编号
     * value 转向实时数据
     *
     * @return
     */
//    public Map<String, List<CrossTurnDataRealtimeDTO>> listTurnDataRealtime(List<CrossDataRealtimePO> abnormalCrossList) {
//        Map<String, List<CrossTurnDataRealtimeDTO>> turnDataRealtimeDTOList = new HashMap<>();
//
//        abnormalCrossList.forEach(entry -> {
//            List<CrossTurnDataRealtimeDTO> crossTurnDataRealtimeDTOS = new ArrayList<>();
//            String crossId = entry.getCrossId();
//            List<CrossTurnInfoDTO> turnList = entry.getTurnList();
//            turnList.forEach(crossTurnInfoDTO -> {
//                CrossTurnDataRealtimeDTO crossTurnDataRealtimeDTO = new CrossTurnDataRealtimeDTO();
//                BeanUtils.copyProperties(entry, crossTurnDataRealtimeDTO);
//                crossTurnDataRealtimeDTO.setTurnType(crossTurnInfoDTO.getTurnType());
//                crossTurnDataRealtimeDTO.setInDir(crossTurnInfoDTO.getInDir());
//                crossTurnDataRealtimeDTO.setOutDir(crossTurnInfoDTO.getOutDir());
//                crossTurnDataRealtimeDTOS.add(crossTurnDataRealtimeDTO);
//            });
//            turnDataRealtimeDTOList.put(crossId, crossTurnDataRealtimeDTOS);
//        });
//
//        return turnDataRealtimeDTOList;
//    }


    /**
     * 计算路口各转向所需的通行时长（秒）
     * 相位的通行时长需大于等于最小绿灯时长，并且小于等于最大绿灯时长
     * 车头间距（米）：车间据如果不在20和7之间取9（默认9）
     * 车头时距（秒）：车时距如果不在2和5之间取2.8（默认2.8）
     * 排队车辆 = 排队长度 / 车头间距
     * 通行时长 = 排队车辆 * 车头时距
     */
    public Double calPassTime(CrossTurnDataRealtimeDTO entity) {

        // todo  优化参数，策略场景优化需要配置参数选择优化方法，读取不同配置优化信息
        Double currentVehheadDist = entity.getVehheadDist();
        if (Objects.isNull(currentVehheadDist) || currentVehheadDist > maxVehheadDist || currentVehheadDist < minVehheadDist) {
            currentVehheadDist = defaultVehheadDist;
        }
        Double currentVehheadTime = entity.getVehheadTime();
        if (Objects.isNull(currentVehheadTime) || currentVehheadTime > maxVehheadTime || currentVehheadTime < minVehheadTime) {
            currentVehheadTime = defaultVehheadTime;
        }
        // 排队车辆
        Double queuedVehicles = null;
        if (Objects.nonNull(entity.getQueueLength())) {
            queuedVehicles = entity.getQueueLength() / currentVehheadDist;
        }
        // 通行时长
        Double calPassTime = queuedVehicles * currentVehheadTime;
        return calPassTime;
    }
}
