package net.wanji.opt.service.impl;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.util.ObjectUtil;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import lombok.extern.slf4j.Slf4j;
import net.wanji.common.dto.CrossIdDTO;
import net.wanji.common.enums.CrossInOutEnum;
import net.wanji.common.utils.tool.CrossUtil;
import net.wanji.databus.bo.CrossDirDataHistAvgBO;
import net.wanji.databus.dao.mapper.CrossDataRealtimeMapper;
import net.wanji.databus.dao.mapper.CrossDirDataHistMapper;
import net.wanji.databus.dao.mapper.CrossDirDataRealtimeMapper;
import net.wanji.databus.dao.mapper.CrossInfoMapper;
import net.wanji.databus.po.CrossDataRealtimePO;
import net.wanji.databus.po.CrossDirDataRealtimePO;
import net.wanji.databus.po.CrossInfoPO;
import net.wanji.opt.dao.mapper.CrossSchemeOptLogMapper;
import net.wanji.opt.dto.CrossIdAndMinutesDTO;
import net.wanji.opt.dto.OptDataExtend;
import net.wanji.opt.po.base.CrossSchemeOptLogPO;
import net.wanji.opt.service.EvaluateService;
import net.wanji.opt.vo.EvaluateCrossDetailVO;
import net.wanji.opt.vo.EvaluateMetricsVO;
import org.jetbrains.annotations.NotNull;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * @author Kent HAN
 * @date 2023/3/10 10:47
 */
@Service
@Slf4j
public class EvaluateServiceImpl implements EvaluateService {
    private final CrossInfoMapper crossInfoMapper;
    private final CrossSchemeOptLogMapper crossSchemeOptLogMapper;
    private final CrossDataRealtimeMapper crossDataRealtimeMapper;
    private final CrossDirDataRealtimeMapper crossDirDataRealtimeMapper;
    private final CrossDirDataHistMapper crossDirDataHistMapper;

    private Gson gson = new GsonBuilder().setDateFormat("yyyy-MM-dd HH:mm:ss").create();

    public EvaluateServiceImpl(CrossInfoMapper crossInfoMapper, CrossSchemeOptLogMapper crossSchemeOptLogMapper,
                               CrossDataRealtimeMapper crossDataRealtimeMapper,
                               CrossDirDataRealtimeMapper crossDirDataRealtimeMapper,
                               CrossDirDataHistMapper crossDirDataHistMapper) {
        this.crossInfoMapper = crossInfoMapper;
        this.crossSchemeOptLogMapper = crossSchemeOptLogMapper;
        this.crossDataRealtimeMapper = crossDataRealtimeMapper;
        this.crossDirDataRealtimeMapper = crossDirDataRealtimeMapper;
        this.crossDirDataHistMapper = crossDirDataHistMapper;
    }

    @Override
    public EvaluateCrossDetailVO evaluateCrossDetail(CrossIdDTO crossIdDTO) {
        EvaluateCrossDetailVO evaluateCrossDetailVO = new EvaluateCrossDetailVO();
        String crossId = crossIdDTO.getCrossId();
        CrossInfoPO crossInfoPO = crossInfoMapper.selectById(crossId);
        String crossName = crossInfoPO.getName();
        evaluateCrossDetailVO.setCrossName(crossName);

        fillOptData(evaluateCrossDetailVO, crossId);
        fillRealtimeData(evaluateCrossDetailVO, crossId);

        return evaluateCrossDetailVO;
    }

    @Override
    public EvaluateMetricsVO evaluateMetrics(CrossIdAndMinutesDTO crossIdAndMinutesDTO) {
        String crossId = crossIdAndMinutesDTO.getCrossId();
        Integer minutes = crossIdAndMinutesDTO.getMinutes();
        EvaluateMetricsVO evaluateMetricsVO = new EvaluateMetricsVO();

        // 路口流量
        List<EvaluateMetricsVO.CrossFlowElement> crossFlow = buildCrossFlow(crossId, minutes);
        evaluateMetricsVO.setCrossFlow(crossFlow);
        // 路口排队长度
        List<EvaluateMetricsVO.QueueLengthElement> queueLength = buildQueueLength(crossId, minutes);
        evaluateMetricsVO.setQueueLength(queueLength);
        // 拥堵指数
        List<EvaluateMetricsVO.TrafficIndexElement> trafficIndex = buildTrafficIndex(crossId, minutes);
        evaluateMetricsVO.setTrafficIndex(trafficIndex);
        // 溢流率
        List<EvaluateMetricsVO.EffusionRateElement> effusionRate = buildEffusionRate(crossId, minutes);
        evaluateMetricsVO.setEffusionRate(effusionRate);

        return evaluateMetricsVO;
    }

    private List<EvaluateMetricsVO.EffusionRateElement> buildEffusionRate(String crossId, Integer minutes) {
        List<EvaluateMetricsVO.EffusionRateElement> res = new ArrayList<>();
        // 实时数据
        Map<Integer, List<CrossDirDataRealtimePO>> dirObjMapRealtime = getRealTimeData(crossId);
        // 历史数据
        Map<Integer, List<CrossDirDataHistAvgBO>> dirObjMapHist = getHistData(crossId, minutes);

        for (Integer dir : dirObjMapRealtime.keySet()) {
            EvaluateMetricsVO.EffusionRateElement effusionRateElement = new EvaluateMetricsVO.EffusionRateElement();
            effusionRateElement.setDir(dir);
            List<CrossDirDataRealtimePO> realtimeObj = dirObjMapRealtime.get(dir);
            if (CollectionUtil.isNotEmpty(realtimeObj)) {
                CrossDirDataRealtimePO crossDirDataRealtimePO = realtimeObj.get(0);
                Double effusionTime = crossDirDataRealtimePO.getEffusionTime();
                Double effusionRate = effusionTime / minutes;
                effusionRateElement.setCurrentRate(effusionRate);
            }
            List<CrossDirDataHistAvgBO> histObj = dirObjMapHist.get(dir);
            if (CollectionUtil.isNotEmpty(histObj)) {
                CrossDirDataHistAvgBO crossDirDataHistAvgBO = histObj.get(0);
                Double avgEffusionTime = crossDirDataHistAvgBO.getAvgEffusionTime();
                Double avgEffusionRate = avgEffusionTime / minutes;
                effusionRateElement.setCompareRate(avgEffusionRate);
            }
            res.add(effusionRateElement);
        }

        return res;
    }

    private List<EvaluateMetricsVO.TrafficIndexElement> buildTrafficIndex(String crossId, Integer minutes) {
        List<EvaluateMetricsVO.TrafficIndexElement> res = new ArrayList<>();
        // 实时数据
        Map<Integer, List<CrossDirDataRealtimePO>> dirObjMapRealtime = getRealTimeData(crossId);
        // 历史数据
        Map<Integer, List<CrossDirDataHistAvgBO>> dirObjMapHist = getHistData(crossId, minutes);

        for (Integer dir : dirObjMapRealtime.keySet()) {
            EvaluateMetricsVO.TrafficIndexElement trafficIndexElement = new EvaluateMetricsVO.TrafficIndexElement();
            trafficIndexElement.setDir(dir);
            List<CrossDirDataRealtimePO> realtimeObj = dirObjMapRealtime.get(dir);
            if (CollectionUtil.isNotEmpty(realtimeObj)) {
                CrossDirDataRealtimePO crossDirDataRealtimePO = realtimeObj.get(0);
                trafficIndexElement.setCurrentIndex(crossDirDataRealtimePO.getTrafficIndex());
            }
            List<CrossDirDataHistAvgBO> histObj = dirObjMapHist.get(dir);
            if (CollectionUtil.isNotEmpty(histObj)) {
                CrossDirDataHistAvgBO crossDirDataHistAvgBO = histObj.get(0);
                trafficIndexElement.setCompareIndex(crossDirDataHistAvgBO.getAvgIndex());
            }
            res.add(trafficIndexElement);
        }

        return res;
    }

    private List<EvaluateMetricsVO.QueueLengthElement> buildQueueLength(String crossId, Integer minutes) {
        List<EvaluateMetricsVO.QueueLengthElement> res = new ArrayList<>();
        // 实时数据
        Map<Integer, List<CrossDirDataRealtimePO>> dirObjMapRealtime = getRealTimeData(crossId);
        // 历史数据
        Map<Integer, List<CrossDirDataHistAvgBO>> dirObjMapHist = getHistData(crossId, minutes);

        for (Integer dir : dirObjMapRealtime.keySet()) {
            EvaluateMetricsVO.QueueLengthElement queueLengthElement = new EvaluateMetricsVO.QueueLengthElement();
            queueLengthElement.setDir(dir);
            List<CrossDirDataRealtimePO> realtimeObj = dirObjMapRealtime.get(dir);
            if (CollectionUtil.isNotEmpty(realtimeObj)) {
                CrossDirDataRealtimePO crossDirDataRealtimePO = realtimeObj.get(0);
                queueLengthElement.setCurrentLength(crossDirDataRealtimePO.getLength());
            }
            List<CrossDirDataHistAvgBO> histObj = dirObjMapHist.get(dir);
            if (CollectionUtil.isNotEmpty(histObj)) {
                CrossDirDataHistAvgBO crossDirDataHistAvgBO = histObj.get(0);
                queueLengthElement.setCompareLength(crossDirDataHistAvgBO.getAvgQueueLength());
            }
            res.add(queueLengthElement);
        }

        return res;
    }

    private List<EvaluateMetricsVO.CrossFlowElement> buildCrossFlow(String crossId, Integer minutes) {
        List<EvaluateMetricsVO.CrossFlowElement> res = new ArrayList<>();
        // 实时数据
        Map<Integer, List<CrossDirDataRealtimePO>> dirObjMapRealtime = getRealTimeData(crossId);
        // 历史数据
        Map<Integer, List<CrossDirDataHistAvgBO>> dirObjMapHist = getHistData(crossId, minutes);

        for (Integer dir : dirObjMapRealtime.keySet()) {
            EvaluateMetricsVO.CrossFlowElement crossFlowElement = new EvaluateMetricsVO.CrossFlowElement();
            crossFlowElement.setDir(dir);
            List<CrossDirDataRealtimePO> realtimeObj = dirObjMapRealtime.get(dir);
            if (CollectionUtil.isNotEmpty(realtimeObj)) {
                CrossDirDataRealtimePO crossDirDataRealtimePO = realtimeObj.get(0);
                crossFlowElement.setCurrentCapacity(crossDirDataRealtimePO.getCapacity());
                crossFlowElement.setCurrentFlow(crossDirDataRealtimePO.getFlow());
            }
            List<CrossDirDataHistAvgBO> histObj = dirObjMapHist.get(dir);
            if (CollectionUtil.isNotEmpty(histObj)) {
                CrossDirDataHistAvgBO crossDirDataHistAvgBO = histObj.get(0);
                crossFlowElement.setCompareCapacity(crossDirDataHistAvgBO.getAvgCapacity());
                crossFlowElement.setCompareFlow(crossDirDataHistAvgBO.getAvgFlow());
            }
            res.add(crossFlowElement);
        }

        return res;
    }

    @NotNull
    private Map<Integer, List<CrossDirDataHistAvgBO>> getHistData(String crossId, Integer minutes) {
        // 获取当前时间之前某分钟的10位时间戳
        long currentSeconds = DateUtil.currentSeconds();
        // todo 测试用 固定当前时间为1676085300
        currentSeconds = 1676085300;
        long preSeconds = currentSeconds - minutes * 60;
        List<CrossDirDataHistAvgBO> crossDirDataHistAvgBOList = crossDirDataHistMapper.selectByCrossIdInOutTimestamp(
                crossId, CrossInOutEnum.IN.getCode(), preSeconds);
        Map<Integer, List<CrossDirDataHistAvgBO>> dirObjMapHist = crossDirDataHistAvgBOList.stream()
                .collect(Collectors.groupingBy(CrossDirDataHistAvgBO::getDirType));
        return dirObjMapHist;
    }

    @NotNull
    private Map<Integer, List<CrossDirDataRealtimePO>> getRealTimeData(String crossId) {
        List<CrossDirDataRealtimePO> crossDirDataRealtimePOList =
                crossDirDataRealtimeMapper.selectByCrossIdAndInOutType(crossId, CrossInOutEnum.IN.getCode());
        Map<Integer, List<CrossDirDataRealtimePO>> dirObjMapRealtime = crossDirDataRealtimePOList.stream()
                .collect(Collectors.groupingBy(CrossDirDataRealtimePO::getDirType));
        return dirObjMapRealtime;
    }

    private void fillRealtimeData(EvaluateCrossDetailVO evaluateCrossDetailVO, String crossId) {
        CrossDataRealtimePO crossDataRealtimePO = crossDataRealtimeMapper.selectByCrossId(crossId);
        evaluateCrossDetailVO.setSceneStartTime(crossDataRealtimePO.getStartTime());
        evaluateCrossDetailVO.setDuration(crossDataRealtimePO.getDuration());
        // 计算服务水平
        Double sturation = crossDataRealtimePO.getSturation();
        String serviceLevel = CrossUtil.getServiceLevel(sturation);
        evaluateCrossDetailVO.setServiceLevel(serviceLevel);
        evaluateCrossDetailVO.setFlow(crossDataRealtimePO.getFlow());
        evaluateCrossDetailVO.setLength(crossDataRealtimePO.getQueueLength());
        evaluateCrossDetailVO.setStopTimes(crossDataRealtimePO.getStopTimes());
    }

    private void fillOptData(EvaluateCrossDetailVO evaluateCrossDetailVO, String crossId) {
        // 选择数据批次时间最大的记录
        List<CrossSchemeOptLogPO> maxList = crossSchemeOptLogMapper.selectMaxByCrossId(crossId);
        if (CollectionUtil.isNotEmpty(maxList)) {
            CrossSchemeOptLogPO crossSchemeOptLogPO = maxList.get(0);
            // 解析dataExtend
            String dataExtend = crossSchemeOptLogPO.getDataExtend();
            OptDataExtend optDataExtend = gson.fromJson(dataExtend, OptDataExtend.class);
            evaluateCrossDetailVO.setSceneName(optDataExtend.getSceneName());
            evaluateCrossDetailVO.setStrategyName(optDataExtend.getStrategyName());
            evaluateCrossDetailVO.setIdeaName(optDataExtend.getIdeaName());
            evaluateCrossDetailVO.setOptStartTime(crossSchemeOptLogPO.getStartTime());
            Date endTime = crossSchemeOptLogPO.getEndTime();
            if (ObjectUtil.isEmpty(endTime)) {
                evaluateCrossDetailVO.setOptStatus(2);
            } else {
                evaluateCrossDetailVO.setOptStatus(1);
            }
        }
    }
}
